@@ -1,8 +1,7 @@
-#!/usr/bin/env python
-# -*- coding: utf-8 -*-
+# coding: utf-8
 #
 # Copyright 2011 Yesudeep Mangalapilly <yesudeep@gmail.com>
-# Copyright 2012 Google, Inc.
+# Copyright 2012 Google, Inc & contributors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -20,6 +19,7 @@
 :module: watchdog.events
 :synopsis: File system events and event handlers.
 :author: yesudeep@google.com (Yesudeep Mangalapilly)
+:author: contact@tiger-222.fr (Mickaël Schoentgen)
 
 Event Classes
 -------------
@@ -88,9 +88,7 @@ Event Handler Classes
 import os.path
 import logging
 import re
-from pathtools.patterns import match_any_paths
-from watchdog.utils import has_attribute
-from watchdog.utils import unicode_paths
+from watchdog.utils.patterns import match_any_paths
 
 
 EVENT_TYPE_MOVED = 'moved'
@@ -99,7 +97,7 @@ EVENT_TYPE_CREATED = 'created'
 EVENT_TYPE_MODIFIED = 'modified'
 
 
-class FileSystemEvent(object):
+class FileSystemEvent:
     """
     Immutable type that represents a file system event that is triggered
     when a change occurs on the monitored file system.
@@ -114,6 +112,14 @@ class FileSystemEvent(object):
     is_directory = False
     """True if event was emitted for a directory; False otherwise."""
 
+    is_synthetic = False
+    """
+    True if event was synthesized; False otherwise.
+
+    These are events that weren't actually broadcast by the OS, but
+    are presumed to have happened based on other, actual events.
+    """
+
     def __init__(self, src_path):
         self._src_path = src_path
 
@@ -158,7 +164,7 @@ class FileSystemMovedEvent(FileSystemEvent):
     event_type = EVENT_TYPE_MOVED
 
     def __init__(self, src_path, dest_path):
-        super(FileSystemMovedEvent, self).__init__(src_path)
+        super().__init__(src_path)
         self._dest_path = dest_path
 
     @property
@@ -189,56 +195,22 @@ class FileDeletedEvent(FileSystemEvent):
 
     event_type = EVENT_TYPE_DELETED
 
-    def __init__(self, src_path):
-        super(FileDeletedEvent, self).__init__(src_path)
-
-    def __repr__(self):
-        return "<%(class_name)s: src_path=%(src_path)r>" %\
-               dict(class_name=self.__class__.__name__,
-                    src_path=self.src_path)
-
 
 class FileModifiedEvent(FileSystemEvent):
     """File system event representing file modification on the file system."""
 
     event_type = EVENT_TYPE_MODIFIED
 
-    def __init__(self, src_path):
-        super(FileModifiedEvent, self).__init__(src_path)
-
-    def __repr__(self):
-        return ("<%(class_name)s: src_path=%(src_path)r>"
-                ) % (dict(class_name=self.__class__.__name__,
-                          src_path=self.src_path))
-
 
 class FileCreatedEvent(FileSystemEvent):
     """File system event representing file creation on the file system."""
 
     event_type = EVENT_TYPE_CREATED
 
-    def __init__(self, src_path):
-        super(FileCreatedEvent, self).__init__(src_path)
-
-    def __repr__(self):
-        return ("<%(class_name)s: src_path=%(src_path)r>"
-                ) % (dict(class_name=self.__class__.__name__,
-                          src_path=self.src_path))
-
 
 class FileMovedEvent(FileSystemMovedEvent):
     """File system event representing file movement on the file system."""
 
-    def __init__(self, src_path, dest_path):
-        super(FileMovedEvent, self).__init__(src_path, dest_path)
-
-    def __repr__(self):
-        return ("<%(class_name)s: src_path=%(src_path)r, "
-                "dest_path=%(dest_path)r>"
-                ) % (dict(class_name=self.__class__.__name__,
-                          src_path=self.src_path,
-                          dest_path=self.dest_path))
-
 
 # Directory events.
 
@@ -249,14 +221,6 @@ class DirDeletedEvent(FileSystemEvent):
     event_type = EVENT_TYPE_DELETED
     is_directory = True
 
-    def __init__(self, src_path):
-        super(DirDeletedEvent, self).__init__(src_path)
-
-    def __repr__(self):
-        return ("<%(class_name)s: src_path=%(src_path)r>"
-                ) % (dict(class_name=self.__class__.__name__,
-                          src_path=self.src_path))
-
 
 class DirModifiedEvent(FileSystemEvent):
     """
@@ -266,14 +230,6 @@ class DirModifiedEvent(FileSystemEvent):
     event_type = EVENT_TYPE_MODIFIED
     is_directory = True
 
-    def __init__(self, src_path):
-        super(DirModifiedEvent, self).__init__(src_path)
-
-    def __repr__(self):
-        return ("<%(class_name)s: src_path=%(src_path)r>"
-                ) % (dict(class_name=self.__class__.__name__,
-                          src_path=self.src_path))
-
 
 class DirCreatedEvent(FileSystemEvent):
     """File system event representing directory creation on the file system."""
@@ -281,32 +237,14 @@ class DirCreatedEvent(FileSystemEvent):
     event_type = EVENT_TYPE_CREATED
     is_directory = True
 
-    def __init__(self, src_path):
-        super(DirCreatedEvent, self).__init__(src_path)
-
-    def __repr__(self):
-        return ("<%(class_name)s: src_path=%(src_path)r>"
-                ) % (dict(class_name=self.__class__.__name__,
-                          src_path=self.src_path))
-
 
 class DirMovedEvent(FileSystemMovedEvent):
     """File system event representing directory movement on the file system."""
 
     is_directory = True
 
-    def __init__(self, src_path, dest_path):
-        super(DirMovedEvent, self).__init__(src_path, dest_path)
 
-    def __repr__(self):
-        return ("<%(class_name)s: src_path=%(src_path)r, "
-                "dest_path=%(dest_path)r>"
-                ) % (dict(class_name=self.__class__.__name__,
-                          src_path=self.src_path,
-                          dest_path=self.dest_path))
-
-
-class FileSystemEventHandler(object):
+class FileSystemEventHandler:
     """
     Base file system event handler that you can override methods from.
     """
@@ -320,14 +258,12 @@ class FileSystemEventHandler(object):
             :class:`FileSystemEvent`
         """
         self.on_any_event(event)
-        _method_map = {
-            EVENT_TYPE_MODIFIED: self.on_modified,
-            EVENT_TYPE_MOVED: self.on_moved,
+        {
             EVENT_TYPE_CREATED: self.on_created,
             EVENT_TYPE_DELETED: self.on_deleted,
-        }
-        event_type = event.event_type
-        _method_map[event_type](event)
+            EVENT_TYPE_MODIFIED: self.on_modified,
+            EVENT_TYPE_MOVED: self.on_moved,
+        }[event.event_type](event)
 
     def on_any_event(self, event):
         """Catch-all event handler.
@@ -382,7 +318,7 @@ class PatternMatchingEventHandler(FileSystemEventHandler):
 
     def __init__(self, patterns=None, ignore_patterns=None,
                  ignore_directories=False, case_sensitive=False):
-        super(PatternMatchingEventHandler, self).__init__()
+        super().__init__()
 
         self._patterns = patterns
         self._ignore_patterns = ignore_patterns
@@ -434,24 +370,16 @@ class PatternMatchingEventHandler(FileSystemEventHandler):
             return
 
         paths = []
-        if has_attribute(event, 'dest_path'):
-            paths.append(unicode_paths.decode(event.dest_path))
+        if hasattr(event, 'dest_path'):
+            paths.append(os.fsdecode(event.dest_path))
         if event.src_path:
-            paths.append(unicode_paths.decode(event.src_path))
+            paths.append(os.fsdecode(event.src_path))
 
         if match_any_paths(paths,
                            included_patterns=self.patterns,
                            excluded_patterns=self.ignore_patterns,
                            case_sensitive=self.case_sensitive):
-            self.on_any_event(event)
-            _method_map = {
-                EVENT_TYPE_MODIFIED: self.on_modified,
-                EVENT_TYPE_MOVED: self.on_moved,
-                EVENT_TYPE_CREATED: self.on_created,
-                EVENT_TYPE_DELETED: self.on_deleted,
-            }
-            event_type = event.event_type
-            _method_map[event_type](event)
+            super().dispatch(event)
 
 
 class RegexMatchingEventHandler(FileSystemEventHandler):
@@ -459,10 +387,14 @@ class RegexMatchingEventHandler(FileSystemEventHandler):
     Matches given regexes with file paths associated with occurring events.
     """
 
-    def __init__(self, regexes=[r".*"], ignore_regexes=[],
+    def __init__(self, regexes=None, ignore_regexes=None,
                  ignore_directories=False, case_sensitive=False):
-        super(RegexMatchingEventHandler, self).__init__()
+        super().__init__()
 
+        if regexes is None:
+            regexes = [r".*"]
+        if ignore_regexes is None:
+            ignore_regexes = []
         if case_sensitive:
             self._regexes = [re.compile(r) for r in regexes]
             self._ignore_regexes = [re.compile(r) for r in ignore_regexes]
@@ -517,60 +449,50 @@ class RegexMatchingEventHandler(FileSystemEventHandler):
             return
 
         paths = []
-        if has_attribute(event, 'dest_path'):
-            paths.append(unicode_paths.decode(event.dest_path))
+        if hasattr(event, 'dest_path'):
+            paths.append(os.fsdecode(event.dest_path))
         if event.src_path:
-            paths.append(unicode_paths.decode(event.src_path))
+            paths.append(os.fsdecode(event.src_path))
 
         if any(r.match(p) for r in self.ignore_regexes for p in paths):
             return
 
         if any(r.match(p) for r in self.regexes for p in paths):
-            self.on_any_event(event)
-            _method_map = {
-                EVENT_TYPE_MODIFIED: self.on_modified,
-                EVENT_TYPE_MOVED: self.on_moved,
-                EVENT_TYPE_CREATED: self.on_created,
-                EVENT_TYPE_DELETED: self.on_deleted,
-            }
-            event_type = event.event_type
-            _method_map[event_type](event)
+            super().dispatch(event)
 
 
 class LoggingEventHandler(FileSystemEventHandler):
     """Logs all the events captured."""
 
+    def __init__(self, logger=None):
+        super().__init__()
+
+        self.logger = logger or logging.root
+
     def on_moved(self, event):
-        super(LoggingEventHandler, self).on_moved(event)
+        super().on_moved(event)
 
         what = 'directory' if event.is_directory else 'file'
-        logging.info("Moved %s: from %s to %s", what, event.src_path,
-                     event.dest_path)
+        self.logger.info("Moved %s: from %s to %s", what, event.src_path,
+                         event.dest_path)
 
     def on_created(self, event):
-        super(LoggingEventHandler, self).on_created(event)
+        super().on_created(event)
 
         what = 'directory' if event.is_directory else 'file'
-        logging.info("Created %s: %s", what, event.src_path)
+        self.logger.info("Created %s: %s", what, event.src_path)
 
     def on_deleted(self, event):
-        super(LoggingEventHandler, self).on_deleted(event)
+        super().on_deleted(event)
 
         what = 'directory' if event.is_directory else 'file'
-        logging.info("Deleted %s: %s", what, event.src_path)
+        self.logger.info("Deleted %s: %s", what, event.src_path)
 
     def on_modified(self, event):
-        super(LoggingEventHandler, self).on_modified(event)
+        super().on_modified(event)
 
         what = 'directory' if event.is_directory else 'file'
-        logging.info("Modified %s: %s", what, event.src_path)
-
-
-class LoggingFileSystemEventHandler(LoggingEventHandler):
-    """
-    For backwards-compatibility. Please use :class:`LoggingEventHandler`
-    instead.
-    """
+        self.logger.info("Modified %s: %s", what, event.src_path)
 
 
 def generate_sub_moved_events(src_dir_path, dest_dir_path):
@@ -590,11 +512,15 @@ def generate_sub_moved_events(src_dir_path, dest_dir_path):
         for directory in directories:
             full_path = os.path.join(root, directory)
             renamed_path = full_path.replace(dest_dir_path, src_dir_path) if src_dir_path else None
-            yield DirMovedEvent(renamed_path, full_path)
+            event = DirMovedEvent(renamed_path, full_path)
+            event.is_synthetic = True
+            yield event
         for filename in filenames:
             full_path = os.path.join(root, filename)
             renamed_path = full_path.replace(dest_dir_path, src_dir_path) if src_dir_path else None
-            yield FileMovedEvent(renamed_path, full_path)
+            event = FileMovedEvent(renamed_path, full_path)
+            event.is_synthetic = True
+            yield event
 
 
 def generate_sub_created_events(src_dir_path):
@@ -610,6 +536,10 @@ def generate_sub_created_events(src_dir_path):
     """
     for root, directories, filenames in os.walk(src_dir_path):
         for directory in directories:
-            yield DirCreatedEvent(os.path.join(root, directory))
+            event = DirCreatedEvent(os.path.join(root, directory))
+            event.is_synthetic = True
+            yield event
         for filename in filenames:
-            yield FileCreatedEvent(os.path.join(root, filename))
\ No newline at end of file
+            event = FileCreatedEvent(os.path.join(root, filename))
+            event.is_synthetic = True
+            yield event
