@@ -1,8 +1,7 @@
-#!/usr/bin/env python
-# -*- coding: utf-8 -*-
+# coding: utf-8
 #
 # Copyright 2011 Yesudeep Mangalapilly <yesudeep@gmail.com>
-# Copyright 2012 Google, Inc.
+# Copyright 2012 Google, Inc & contributors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -19,45 +18,31 @@
 """
 :module: watchdog.watchmedo
 :author: yesudeep@google.com (Yesudeep Mangalapilly)
+:author: contact@tiger-222.fr (Mickaël Schoentgen)
 :synopsis: ``watchmedo`` shell script utility.
 """
 
+import errno
+import os
 import os.path
 import sys
 import yaml
 import time
 import logging
+from io import StringIO
 
-try:
-    from cStringIO import StringIO
-except ImportError:
-    try:
-        from StringIO import StringIO
-    except ImportError:
-        from io import StringIO
-
-from argh import arg, alias, ArghParser
-
+from argh import arg, aliases, ArghParser, expects_obj
 from watchdog.version import VERSION_STRING
-
-from watchdog.utils import (
-    read_text_file,
-    load_class
-)
-
-from pathtools.path import absolute_path, parent_dir_path
+from watchdog.utils import WatchdogShutdown, load_class
 
 
-logging.basicConfig(level=logging.DEBUG)
+logging.basicConfig(level=logging.INFO)
 
-CURRENT_DIR = absolute_path(os.getcwd())
-DEFAULT_TRICKS_FILE_NAME = 'tricks.yaml'
-DEFAULT_TRICKS_FILE_PATH = os.path.join(CURRENT_DIR, DEFAULT_TRICKS_FILE_NAME)
 CONFIG_KEY_TRICKS = 'tricks'
 CONFIG_KEY_PYTHON_PATH = 'python-path'
 
 
-def path_split(pathname_spec, separator=os.path.sep):
+def path_split(pathname_spec, separator=os.pathsep):
     """
     Splits a pathname specification separated by an OS-dependent separator.
 
@@ -92,9 +77,8 @@ def load_config(tricks_file_pathname):
     :returns:
         A dictionary of configuration information.
     """
-    content = read_text_file(tricks_file_pathname)
-    config = yaml.load(content)
-    return config
+    with open(tricks_file_pathname, 'rb') as f:
+        return yaml.safe_load(f.read())
 
 
 def parse_patterns(patterns_spec, ignore_patterns_spec, separator=';'):
@@ -128,7 +112,7 @@ def observe_with(observer, event_handler, pathnames, recursive):
     try:
         while True:
             time.sleep(1)
-    except KeyboardInterrupt:
+    except WatchdogShutdown:
         observer.stop()
     observer.join()
 
@@ -151,26 +135,26 @@ def schedule_tricks(observer, tricks, pathname, recursive):
         for name, value in list(trick.items()):
             TrickClass = load_class(name)
             handler = TrickClass(**value)
-            trick_pathname = absolute_path(
-                getattr(handler, 'source_directory', None) or pathname)
+            trick_pathname = getattr(handler, 'source_directory', None) or pathname
             observer.schedule(handler, trick_pathname, recursive)
 
 
-@alias('tricks')
+@aliases('tricks')
 @arg('files',
      nargs='*',
      help='perform tricks from given file')
 @arg('--python-path',
      default='.',
-     help='paths separated by %s to add to the python path' % os.path.sep)
+     help='paths separated by %s to add to the python path' % os.pathsep)
 @arg('--interval',
      '--timeout',
      dest='timeout',
      default=1.0,
-     help='use this as the polling interval/blocking timeout')
+     help='use this as the polling interval/blocking timeout (in seconds)')
 @arg('--recursive',
      default=True,
      help='recursively monitor paths')
+@expects_obj
 def tricks_from(args):
     """
     Subcommand to execute tricks from a tricks configuration file.
@@ -186,20 +170,22 @@ def tricks_from(args):
         observer = Observer(timeout=args.timeout)
 
         if not os.path.exists(tricks_file):
-            raise IOError("cannot find tricks file: %s" % tricks_file)
+            raise OSError(errno.ENOENT, os.strerror(errno.ENOENT), tricks_file)
 
         config = load_config(tricks_file)
 
         try:
             tricks = config[CONFIG_KEY_TRICKS]
         except KeyError:
-            raise KeyError("No `%s' key specified in %s." % (
+            raise KeyError("No %r key specified in %s." % (
                            CONFIG_KEY_TRICKS, tricks_file))
 
         if CONFIG_KEY_PYTHON_PATH in config:
             add_to_sys_path(config[CONFIG_KEY_PYTHON_PATH])
 
-        dir_path = parent_dir_path(tricks_file)
+        dir_path = os.path.dirname(tricks_file)
+        if not dir_path:
+            dir_path = os.path.relpath(os.getcwd())
         schedule_tricks(observer, tricks, dir_path, args.recursive)
         observer.start()
         observers.append(observer)
@@ -207,7 +193,7 @@ def tricks_from(args):
     try:
         while True:
             time.sleep(1)
-    except KeyboardInterrupt:
+    except WatchdogShutdown:
         for o in observers:
             o.unschedule_all()
             o.stop()
@@ -215,13 +201,13 @@ def tricks_from(args):
         o.join()
 
 
-@alias('generate-tricks-yaml')
+@aliases('generate-tricks-yaml')
 @arg('trick_paths',
      nargs='*',
      help='Dotted paths for all the tricks you want to generate')
 @arg('--python-path',
      default='.',
-     help='paths separated by %s to add to the python path' % os.path.sep)
+     help='paths separated by %s to add to the python path' % os.pathsep)
 @arg('--append-to-file',
      default=None,
      help='appends the generated tricks YAML to a file; \
@@ -232,6 +218,7 @@ if not specified, prints to standard output')
      default=False,
      help='if --append-to-file is not specified, produces output for \
 appending instead of a complete tricks yaml file.')
+@expects_obj
 def tricks_generate_yaml(args):
     """
     Subcommand to generate Yaml configuration for tricks named on the command
@@ -261,9 +248,8 @@ def tricks_generate_yaml(args):
     else:
         if not os.path.exists(args.append_to_file):
             content = header + content
-        output = open(args.append_to_file, 'ab')
-        output.write(content)
-        output.close()
+        with open(args.append_to_file, 'ab') as output:
+            output.write(content)
 
 
 @arg('directories',
@@ -318,6 +304,7 @@ def tricks_generate_yaml(args):
 @arg('--debug-force-inotify',
      default=False,
      help='[debug] forces Linux inotify(7)')
+@expects_obj
 def log(args):
     """
     Subcommand to log file system events to the console.
@@ -351,8 +338,8 @@ def log(args):
     elif args.debug_force_fsevents:
         from watchdog.observers.fsevents import FSEventsObserver as Observer
     else:
-    # Automatically picks the most appropriate observer for the platform
-    # on which it is running.
+        # Automatically picks the most appropriate observer for the platform
+        # on which it is running.
         from watchdog.observers import Observer
     observer = Observer(timeout=args.timeout)
     observe_with(observer, handler, args.directories, args.recursive)
@@ -416,6 +403,16 @@ Example option usage::
      action='store_true',
      default=False,
      help="wait for process to finish to avoid multiple simultaneous instances")
+@arg('-W', '--drop',
+     dest='drop_during_process',
+     action='store_true',
+     default=False,
+     help="Ignore events that occur while command is still being executed "
+          "to avoid multiple simultaneous instances")
+@arg('--debug-force-polling',
+     default=False,
+     help='[debug] forces polling')
+@expects_obj
 def shell_command(args):
     """
     Subcommand to execute shell commands in response to file system events.
@@ -423,19 +420,24 @@ def shell_command(args):
     :param args:
         Command line argument options.
     """
-    from watchdog.observers import Observer
     from watchdog.tricks import ShellCommandTrick
 
     if not args.command:
         args.command = None
 
+    if args.debug_force_polling:
+        from watchdog.observers.polling import PollingObserver as Observer
+    else:
+        from watchdog.observers import Observer
+
     patterns, ignore_patterns = parse_patterns(args.patterns,
                                                args.ignore_patterns)
     handler = ShellCommandTrick(shell_command=args.command,
                                 patterns=patterns,
                                 ignore_patterns=ignore_patterns,
                                 ignore_directories=args.ignore_directories,
-                                wait_for_process=args.wait_for_process)
+                                wait_for_process=args.wait_for_process,
+                                drop_during_process=args.drop_during_process)
     observer = Observer(timeout=args.timeout)
     observe_with(observer, handler, args.directories, args.recursive)
 
@@ -489,11 +491,15 @@ try to interpret them.
      dest='signal',
      default='SIGINT',
      help='stop the subprocess with this signal (default SIGINT)')
+@arg('--debug-force-polling',
+     default=False,
+     help='[debug] forces polling')
 @arg('--kill-after',
      dest='kill_after',
      default=10.0,
      help='when stopping, kill the subprocess after the specified timeout '
           '(default 10)')
+@expects_obj
 def auto_restart(args):
     """
     Subcommand to start a long-running subprocess and restart it
@@ -502,26 +508,36 @@ def auto_restart(args):
     :param args:
         Command line argument options.
     """
-    from watchdog.observers import Observer
+
+    if args.debug_force_polling:
+        from watchdog.observers.polling import PollingObserver as Observer
+    else:
+        from watchdog.observers import Observer
+
     from watchdog.tricks import AutoRestartTrick
     import signal
-    import re
 
     if not args.directories:
         args.directories = ['.']
 
     # Allow either signal name or number.
-    if re.match('^SIG[A-Z]+$', args.signal):
+    if args.signal.startswith("SIG"):
         stop_signal = getattr(signal, args.signal)
     else:
         stop_signal = int(args.signal)
 
-    # Handle SIGTERM in the same manner as SIGINT so that
-    # this program has a chance to stop the child process.
-    def handle_sigterm(_signum, _frame):
-        raise KeyboardInterrupt()
+    # Handle termination signals by raising a semantic exception which will
+    # allow us to gracefully unwind and stop the observer
+    termination_signals = {signal.SIGTERM, signal.SIGINT}
 
-    signal.signal(signal.SIGTERM, handle_sigterm)
+    def handler_termination_signal(_signum, _frame):
+        # Neuter all signals so that we don't attempt a double shutdown
+        for signum in termination_signals:
+            signal.signal(signum, signal.SIG_IGN)
+        raise WatchdogShutdown
+
+    for signum in termination_signals:
+        signal.signal(signum, handler_termination_signal)
 
     patterns, ignore_patterns = parse_patterns(args.patterns,
                                                args.ignore_patterns)
@@ -535,12 +551,16 @@ def auto_restart(args):
                                kill_after=args.kill_after)
     handler.start()
     observer = Observer(timeout=args.timeout)
-    observe_with(observer, handler, args.directories, args.recursive)
-    handler.stop()
+    try:
+        observe_with(observer, handler, args.directories, args.recursive)
+    except WatchdogShutdown:
+        pass
+    finally:
+        handler.stop()
 
 
 epilog = """Copyright 2011 Yesudeep Mangalapilly <yesudeep@gmail.com>.
-Copyright 2012 Google, Inc.
+Copyright 2012 Google, Inc & contributors.
 
 Licensed under the terms of the Apache license, version 2.0. Please see
 LICENSE in the source code for more information."""
